@@ -45,9 +45,18 @@ from py7zr.archiveinfo import Folder, Header, SignatureHeader
 from py7zr.callbacks import ExtractCallback
 from py7zr.compressor import SupportedMethods, get_methods_names
 from py7zr.exceptions import Bad7zFile, CrcError, DecompressionError, InternalError, UnsupportedCompressionMethodError
-from py7zr.helpers import ArchiveTimestamp, MemIO, NullIO, calculate_crc32, filetime_to_dt, readlink
-from py7zr.properties import DEFAULT_FILTERS, MAGIC_7Z, get_default_blocksize, FILTER_DEFLATE64
-from py7zr.win32compat import is_windows_native_python, is_windows_unc_path
+from py7zr.helpers import (
+    ArchiveTimestamp,
+    MemIO,
+    NullIO,
+    calculate_crc32,
+    check_archive_path,
+    filetime_to_dt,
+    get_sanitized_output_path,
+    is_target_path_valid,
+    readlink,
+)
+from py7zr.properties import DEFAULT_FILTERS, FILTER_DEFLATE64, MAGIC_7Z, get_default_blocksize, get_memory_limit
 
 if sys.platform.startswith("win"):
     import _winapi
@@ -559,34 +568,10 @@ class SevenZipFile(contextlib.AbstractContextManager):
                         break
                     i += 1
             fnames.append(outname)
-            # check f.filename has invalid directory traversals
-            if path is None:
-                # do following but is_relative_to introduced in py 3.9
-                # so I replaced it with relative_to. when condition is not satisfied, raise ValueError
-                # if not pathlib.Path(...).joinpath(outname.lstrip("./")).is_relative_to(...):
-                #    raise Bad7zFile
-                try:
-                    pathlib.Path(os.getcwd()).joinpath(outname.lstrip("./")).relative_to(os.getcwd())
-                except ValueError:
-                    raise Bad7zFile
-                outfilename = pathlib.Path(outname.lstrip("./"))
+            if path is None or path.is_absolute():
+                outfilename = get_sanitized_output_path(outname, path)
             else:
-                outfilename = path.joinpath(outname.lstrip("./"))
-                try:
-                    outfilename.relative_to(path)
-                except ValueError:
-                    raise Bad7zFile
-            # When python on Windows and not python on Cygwin,
-            # Add win32 file namespace to exceed microsoft windows
-            # path length limitation to 260 bytes
-            # ref.
-            # https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file
-            # In editions of Windows before Windows 10 version 1607,
-            # the maximum length for a path is MAX_PATH, which is defined as
-            # 260 characters. In later versions of Windows, changing a registry key
-            # or select option when python installation is required to remove the limit.
-            if is_windows_native_python() and outfilename.is_absolute() and not is_windows_unc_path(outfilename):
-                outfilename = pathlib.WindowsPath("\\\\?\\" + str(outfilename))
+                outfilename = get_sanitized_output_path(outname, pathlib.Path(os.getcwd()).joinpath(path))
             if targets is not None and f.filename not in targets:
                 self.worker.register_filelike(f.id, None)
                 continue
@@ -626,12 +611,14 @@ class SevenZipFile(contextlib.AbstractContextManager):
         if callback is not None:
             self.worker.extract(
                 self.fp,
+                path,
                 parallel=(not self.password_protected and not self._filePassed),
                 q=self.q,
             )
         else:
             self.worker.extract(
                 self.fp,
+                path,
                 parallel=(not self.password_protected and not self._filePassed),
             )
 
@@ -969,7 +956,7 @@ class SevenZipFile(contextlib.AbstractContextManager):
     def extractall(self, path: Optional[Any] = None, callback: Optional[ExtractCallback] = None) -> None:
         """Extract all members from the archive to the current working
         directory and set owner, modification time and permissions on
-        directories afterwards. `path' specifies a different directory
+        directories afterwards. ``path`` specifies a different directory
         to extract to.
         """
         self._extract(path=path, return_dict=False, callback=callback)
@@ -1015,19 +1002,18 @@ class SevenZipFile(contextlib.AbstractContextManager):
             raise ValueError("specified path is not a directory or a file")
 
     def write(self, file: Union[pathlib.Path, str], arcname: Optional[str] = None):
-        """Write single target file into archive(Not implemented yet)."""
+        """Write single target file into archive."""
         if isinstance(file, str):
             path = pathlib.Path(file)
         elif isinstance(file, pathlib.Path):
             path = file
         else:
             raise ValueError("Unsupported file type.")
-        self.header.initialize()
+        folder = self.header.initialize()
         file_info = self._make_file_info(path, arcname, self.dereference)
         self.header.files_info.files.append(file_info)
         self.header.files_info.emptyfiles.append(file_info["emptystream"])
         self.files.append(file_info)
-        folder = self.header.main_streams.unpackinfo.folders[-1]
         self.worker.archive(self.fp, self.files, folder, deref=self.dereference)
 
     def writed(self, targets: Dict[str, IO[Any]]) -> None:
@@ -1035,6 +1021,11 @@ class SevenZipFile(contextlib.AbstractContextManager):
             self.writef(input, target)
 
     def writef(self, bio: IO[Any], arcname: str):
+        if not check_archive_path(arcname):
+            raise ValueError(f"Specified path is bad: {arcname}")
+        return self._writef(bio, arcname)
+
+    def _writef(self, bio: IO[Any], arcname: str):
         if isinstance(bio, io.BytesIO):
             size = bio.getbuffer().nbytes
         elif isinstance(bio, io.TextIOBase):
@@ -1051,12 +1042,11 @@ class SevenZipFile(contextlib.AbstractContextManager):
         else:
             raise ValueError("Wrong argument passed for argument bio.")
         if size > 0:
-            self.header.initialize()
+            folder = self.header.initialize()
             file_info = self._make_file_info_from_name(bio, size, arcname)
             self.header.files_info.files.append(file_info)
             self.header.files_info.emptyfiles.append(file_info["emptystream"])
             self.files.append(file_info)
-            folder = self.header.main_streams.unpackinfo.folders[-1]
             self.worker.archive(self.fp, self.files, folder, deref=False)
         else:
             file_info = self._make_file_info_from_name(bio, size, arcname)
@@ -1065,12 +1055,17 @@ class SevenZipFile(contextlib.AbstractContextManager):
             self.files.append(file_info)
 
     def writestr(self, data: Union[str, bytes, bytearray, memoryview], arcname: str):
+        if not check_archive_path(arcname):
+            raise ValueError(f"Specified path is bad: {arcname}")
+        return self._writestr(data, arcname)
+
+    def _writestr(self, data: Union[str, bytes, bytearray, memoryview], arcname: str):
         if not isinstance(arcname, str):
             raise ValueError("Unsupported arcname")
         if isinstance(data, str):
-            self.writef(io.BytesIO(data.encode("UTF-8")), arcname)
+            self._writef(io.BytesIO(data.encode("UTF-8")), arcname)
         elif isinstance(data, bytes) or isinstance(data, bytearray) or isinstance(data, memoryview):
-            self.writef(io.BytesIO(data), arcname)
+            self._writef(io.BytesIO(bytes(data)), arcname)
         else:
             raise ValueError("Unsupported data type.")
 
@@ -1093,7 +1088,9 @@ class SevenZipFile(contextlib.AbstractContextManager):
         self._var_release()
 
     def reset(self) -> None:
-        """When read mode, it reset file pointer, decompress worker and decompressor"""
+        """
+        When read mode, it reset file pointer, decompress worker and decompressor
+        """
         if self.mode == "r":
             self.fp.seek(self.afterheader)
             self.worker = Worker(self.files, self.afterheader, self.header, self.mp)
@@ -1125,7 +1122,9 @@ class SevenZipFile(contextlib.AbstractContextManager):
         for f in self.files:
             self.worker.register_filelike(f.id, None)
         try:
-            self.worker.extract(self.fp, parallel=(not self.password_protected), skip_notarget=False)  # TODO: print progress
+            self.worker.extract(
+                self.fp, None, parallel=(not self.password_protected), skip_notarget=False
+            )  # TODO: print progress
         except CrcError as crce:
             return crce.args[2]
         else:
@@ -1159,14 +1158,18 @@ def is_7zfile(file: Union[BinaryIO, str, pathlib.Path]) -> bool:
 
 
 def unpack_7zarchive(archive, path, extra=None):
-    """Function for registering with shutil.register_unpack_format()"""
+    """
+    Function for registering with shutil.register_unpack_format().
+    """
     arc = SevenZipFile(archive)
     arc.extractall(path)
     arc.close()
 
 
 def pack_7zarchive(base_name, base_dir, owner=None, group=None, dry_run=None, logger=None):
-    """Function for registering with shutil.register_archive_format()"""
+    """
+    Function for registering with shutil.register_archive_format().
+    """
     target_name = "{}.7z".format(base_name)
     with SevenZipFile(target_name, mode="w") as archive:
         archive.writeall(path=base_dir)
@@ -1174,7 +1177,9 @@ def pack_7zarchive(base_name, base_dir, owner=None, group=None, dry_run=None, lo
 
 
 class Worker:
-    """Extract worker class to invoke handler"""
+    """
+    Extract worker class to invoke handler.
+    """
 
     def __init__(self, files, src_start: int, header, mp=False) -> None:
         self.target_filepath: Dict[int, Union[MemIO, pathlib.Path, None]] = {}
@@ -1188,7 +1193,7 @@ class Worker:
         else:
             self.concurrent = Thread
 
-    def extract(self, fp: BinaryIO, parallel: bool, skip_notarget=True, q=None) -> None:
+    def extract(self, fp: BinaryIO, path: Optional[pathlib.Path], parallel: bool, skip_notarget=True, q=None) -> None:
         """Extract worker method to handle 7zip folder and decompress each files."""
         if hasattr(self.header, "main_streams") and self.header.main_streams is not None:
             src_end = self.src_start + self.header.main_streams.packinfo.packpositions[-1]
@@ -1197,6 +1202,7 @@ class Worker:
                 self.extract_single(
                     fp,
                     self.files,
+                    path,
                     self.src_start,
                     src_end,
                     q,
@@ -1207,7 +1213,7 @@ class Worker:
                 positions = self.header.main_streams.packinfo.packpositions
                 empty_files = [f for f in self.files if f.emptystream]
                 if not parallel:
-                    self.extract_single(fp, empty_files, 0, 0, q)
+                    self.extract_single(fp, empty_files, path, 0, 0, q)
                     for i in range(numfolders):
                         if skip_notarget:
                             if not any([self.target_filepath.get(f.id, None) for f in folders[i].files]):
@@ -1215,6 +1221,7 @@ class Worker:
                         self.extract_single(
                             fp,
                             folders[i].files,
+                            path,
                             self.src_start + positions[i],
                             self.src_start + positions[i + 1],
                             q,
@@ -1224,7 +1231,7 @@ class Worker:
                     if getattr(fp, "name", None) is None:
                         raise InternalError("Caught unknown variable status error")
                     filename: str = getattr(fp, "name", "")  # do not become "" but it is for type check.
-                    self.extract_single(open(filename, "rb"), empty_files, 0, 0, q)
+                    self.extract_single(open(filename, "rb"), empty_files, path, 0, 0, q)
                     concurrent_tasks = []
                     exc_q: queue.Queue = queue.Queue()
                     for i in range(numfolders):
@@ -1236,6 +1243,7 @@ class Worker:
                             args=(
                                 filename,
                                 folders[i].files,
+                                path,
                                 self.src_start + positions[i],
                                 self.src_start + positions[i + 1],
                                 q,
@@ -1254,26 +1262,29 @@ class Worker:
                         raise exc_info[1].with_traceback(exc_info[2])
         else:
             empty_files = [f for f in self.files if f.emptystream]
-            self.extract_single(fp, empty_files, 0, 0, q)
+            self.extract_single(fp, empty_files, path, 0, 0, q)
 
     def extract_single(
         self,
         fp: Union[BinaryIO, str],
         files,
+        path,
         src_start: int,
         src_end: int,
         q: Optional[queue.Queue],
         exc_q: Optional[queue.Queue] = None,
         skip_notarget=True,
     ) -> None:
-        """Single thread extractor that takes file lists in single 7zip folder."""
+        """
+        Single thread extractor that takes file lists in single 7zip folder.
+        """
         if files is None:
             return
         try:
             if isinstance(fp, str):
                 fp = open(fp, "rb")
             fp.seek(src_start)
-            self._extract_single(fp, files, src_end, q, skip_notarget)
+            self._extract_single(fp, files, path, src_end, q, skip_notarget)
         except Exception as e:
             if exc_q is None:
                 raise e
@@ -1285,6 +1296,7 @@ class Worker:
         self,
         fp: BinaryIO,
         files,
+        path,
         src_end: int,
         q: Optional[queue.Queue],
         skip_notarget=True,
@@ -1317,20 +1329,28 @@ class Worker:
                         with io.BytesIO() as ofp:
                             self.decompress(fp, f.folder, ofp, f.uncompressed, f.compressed, src_end)
                             dst: str = ofp.read().decode("utf-8")
-                            # fileish.unlink(missing_ok=True) > py3.7
-                            if fileish.exists():
-                                fileish.unlink()
-                            if sys.platform == "win32":  # hint for mypy
-                                _winapi.CreateJunction(str(fileish), dst)  # noqa
+                            if is_target_path_valid(path, fileish.parent.joinpath(dst)):
+                                # fileish.unlink(missing_ok=True) > py3.7
+                                if fileish.exists():
+                                    fileish.unlink()
+                                if sys.platform == "win32":  # hint for mypy
+                                    _winapi.CreateJunction(str(fileish), dst)  # noqa
+                            else:
+                                raise Bad7zFile("Junction point out of target directory.")
                     elif f.is_symlink and not isinstance(fileish, MemIO):
                         with io.BytesIO() as omfp:
                             self.decompress(fp, f.folder, omfp, f.uncompressed, f.compressed, src_end)
                             omfp.seek(0)
-                            sym_target = pathlib.Path(omfp.read().decode("utf-8"))
-                            # fileish.unlink(missing_ok=True) > py3.7
-                            if fileish.exists():
-                                fileish.unlink()
-                            fileish.symlink_to(sym_target)
+                            dst = omfp.read().decode("utf-8")
+                            # check sym_target points inside an archive target?
+                            if is_target_path_valid(path, fileish.parent.joinpath(dst)):
+                                sym_target = pathlib.Path(dst)
+                                # fileish.unlink(missing_ok=True) > py3.7
+                                if fileish.exists():
+                                    fileish.unlink()
+                                fileish.symlink_to(sym_target)
+                            else:
+                                raise Bad7zFile("Symlink point out of target directory.")
                     else:
                         with fileish.open(mode="wb") as obfp:
                             crc32 = self.decompress(fp, f.folder, obfp, f.uncompressed, f.compressed, src_end)
@@ -1351,7 +1371,9 @@ class Worker:
             self._check(fp, just_check, src_end)
 
     def _check(self, fp, check_target, src_end):
-        # delayed execution of crc check.
+        """
+        delayed execution of crc check.
+        """
         for f in check_target:
             with NullIO() as ofp:
                 crc32 = self.decompress(fp, f.folder, ofp, f.uncompressed, f.compressed, src_end)
@@ -1367,7 +1389,8 @@ class Worker:
         compressed_size: Optional[int],
         src_end: int,
     ) -> int:
-        """decompressor wrapper called from extract method.
+        """
+        decompressor wrapper called from extract method.
 
         :parameter fp: archive source file pointer
         :parameter folder: Folder object that have decompressor object.
@@ -1375,14 +1398,17 @@ class Worker:
         :parameter size: uncompressed size of target file.
         :parameter compressed_size: compressed size of target file.
         :parameter src_end: end position of the folder
+
         :returns None
+
         """
         assert folder is not None
         out_remaining = size
+        max_block_size = get_memory_limit()
         crc32 = 0
         decompressor = folder.get_decompressor(compressed_size)
         while out_remaining > 0:
-            tmp = decompressor.decompress(fp, out_remaining)
+            tmp = decompressor.decompress(fp, min(out_remaining, max_block_size))
             if len(tmp) > 0:
                 out_remaining -= len(tmp)
                 fq.write(tmp)
@@ -1395,7 +1421,9 @@ class Worker:
         return crc32
 
     def _find_link_target(self, target):
-        """Find the target member of a symlink or hardlink member in the archive."""
+        """
+        Find the target member of a symlink or hardlink member in the archive.
+        """
         targetname: str = target.as_posix()
         linkname = readlink(targetname)
         # Check windows full path symlinks
@@ -1454,8 +1482,9 @@ class Worker:
                 self.header.files_info.files[self.last_file_index]["maxsize"] = foutsize
         # Update size data in header
         self.header.main_streams.packinfo.numstreams += 1
-        self.header.main_streams.packinfo.crcs.append(compressor.digest)
-        self.header.main_streams.packinfo.digestdefined.append(True)
+        if self.header.main_streams.packinfo.enable_digests:
+            self.header.main_streams.packinfo.crcs.append(compressor.digest)
+            self.header.main_streams.packinfo.digestdefined.append(True)
         self.header.main_streams.packinfo.packsizes.append(compressor.packsize)
         folder.unpacksizes = compressor.unpacksizes
 
